from __future__ import absolute_import, unicode_literals

from tornado import gen
from tornado.concurrent import is_future
import zmq
import zmq.eventloop.ioloop
from zmq.eventloop.zmqstream import ZMQStream

from logdog.core.msg import Msg
from .base import BaseConnector


zmq.eventloop.ioloop.install()


class ZMQTunnel(BaseConnector):
    defaults = BaseConnector.defaults(
        unique=True,
        connect=(),
        bind=(),
        socket=None,
    )

    def __init__(self, *args, **kwargs):
        if not isinstance(kwargs.get('bind', ()), (list, tuple)):
            kwargs['bind'] = (kwargs['bind'],)
        if not isinstance(kwargs.get('connect', ()), (list, tuple)):
            kwargs['connect'] = (kwargs['connect'],)

        self.stream = self.socket = None
        self.ctx = zmq.Context()

        super(ZMQTunnel, self).__init__(*args, **kwargs)

    def __str__(self):
        return 'ZMQ-TUNNEL:{}:{}'.format(self.config.socket,
                                          ','.join(self.config.bind) or
                                          ','.join(self.config.connect) or 'None')

    @classmethod
    def __singleton_key__(cls, passed_args, passed_kwargs):
        key = super(ZMQTunnel, cls).__singleton_key__(passed_args, passed_kwargs)
        connect = passed_kwargs.get('connect', cls.defaults.connect) or ()
        if not isinstance(connect, (list, tuple)):
            connect = [connect]
        bind = passed_kwargs.get('bind', cls.defaults.bind) or ()
        if not isinstance(bind, (list, tuple)):
            bind = [bind]
        return '{key}::socket-type={socket_type}::bind={bind}::connect={connect}'.format(
            key=key,
            socket_type=passed_kwargs.get('socket', cls.defaults.socket),
            connect=','.join(sorted(connect)) or 'None',
            bind=','.join(sorted(bind)) or 'None',
        )

    def _pre_start(self):
        self.socket = self.ctx.socket(getattr(zmq, self.config.socket))

        if 'bind' in self.config:
            for addr in self.config.bind:
                self.socket.bind(addr)

        if 'connect' in self.config:
            for addr in self.config.connect:
                self.socket.connect(addr)

        self.stream = ZMQStream(self.socket, io_loop=self.app.io_loop)
        self.stream.on_recv(self.on_recv)

    @gen.coroutine
    def on_recv(self, data):
        data = Msg.deserialize_jsonb(data[0])
        ret = self._forward(data)
        if is_future(ret):
            yield ret

    def send(self, data):
        if self.started:
            data = data.serialize_jsonb()
            return self.stream.send(msg=data)


from __future__ import absolute_import, unicode_literals

import logging
import os
from tornado import gen
from tornado.web import Application, url

from ..base import BaseViewer


logger = logging.getLogger(__name__)


class WebApp(Application):
    def __init__(self, *args, **kwargs):
        from .web import IndexHandler, LogWebSocket
        handlers = [
            url(r'/', IndexHandler, name='logdog.index'),
            url(r'/ws/logs', LogWebSocket, name='logdog.logs-ws'),
        ]
        self.ws = []
        kwargs['handlers'] = handlers
        super(WebApp, self).__init__(*args, **kwargs)


class WebUI(BaseViewer):
    defaults = BaseViewer.defaults(
        unique=True,
        port=8888,
        address='',
        autoreload=False,
        debug=False,
        static_path=os.path.join(os.path.dirname(__file__), 'static'),
        template_path=os.path.join(os.path.dirname(__file__), 'assets/html'),
    )

    def __init__(self, *args, **kwargs):
        super(WebUI, self).__init__(*args, **kwargs)
        self._web_app = WebApp(**self.config)

    def __str__(self):
        return 'WEBUI:http://{}:{}'.format(self.config.address or '0.0.0.0', self.config.port)

    @classmethod
    def __singleton_key__(cls, passed_args, passed_kwargs):
        key = super(WebUI, cls).__singleton_key__(passed_args, passed_kwargs)
        return '{key}:address={address}:port={port}'.format(
            key=key,
            address=passed_kwargs.get('address', cls.defaults.address),
            port=passed_kwargs.get('port', cls.defaults.port)
        )

    @gen.coroutine
    def _pre_start(self):
        logger.info('[%s] Starting %s:%s...', self, self.config.address or '0.0.0.0', self.config.port)
        try:
            self._web_app.listen(address=self.config.address, port=self.config.port)
        except Exception as e:
            logger.exception(e)

    def on_recv(self, data):
        if self.started:
            for ws in self._web_app.ws:
                ws.write_message(data.serialize_json())



from __future__ import absolute_import, unicode_literals

import logging
from tornado import gen

from logdog.core.base_role import BaseRole


logger = logging.getLogger(__name__)


class BasePipe(BaseRole):
    defaults = BaseRole.defaults.copy_and_update(
        unique=True
    )

    def __init__(self, *args, **kwargs):
        super(BasePipe, self).__init__(*args, **kwargs)
        self.items = [self.construct_subrole(name, conf) for name, conf in self.items]
        self._link_pipe_objects()
        if self.items:
            self.items[-1].set_output(None)

    def __str__(self):
        return 'PIPE:{}'.format(self._oid)

    def _is_valid_pipe(self):
        active_pipe = self.active_items
        in_ = active_pipe[0] if active_pipe else None
        for p in active_pipe[1:]:
            if in_.output is not p:
                logger.warning('[%s] Lost connectivity in the pipe. Check the configuration. %s -x-> %s', self, in_, p)
                return False
            in_ = p
        return True

    @gen.coroutine
    def start(self):
        if not self.started and self._is_valid_pipe():
            logger.debug('[%s] Starting %s', self, ' -> '.join(map(str, self.items)))
            yield [po.start() for po in reversed(self.items)]
            yield super(BasePipe, self).start()

    @gen.coroutine
    def stop(self):
        if self.started:
            yield [po.stop() for po in self.items]
            yield super(BasePipe, self).stop()

    def _link_pipe_objects(self):
        active_pipe = self.active_items
        obj_first = active_pipe[0] if active_pipe else None
        for obj_next in active_pipe[1:]:
            obj_first.set_output(obj_next)
            obj_next.set_input(obj_first)
            obj_first = obj_next

        for obj in reversed(active_pipe):
            obj.link_methods()

    def set_input(self, obj):
        logger.debug('[%s] Pipe initialized with %s.', self, obj)
        if self.active_items:
            self.active_items[0].set_input(obj)

    def set_output(self, obj):
        if self.active_items:
            self.active_items[-1].set_output(obj)

    def _input_forwarder(self, data):
        if self.active_items:
            self._input_forwarder = self.active_items[0].send
            return self._input_forwarder(data)


from __future__ import absolute_import, unicode_literals

import re
from .base import BaseParser


class Regex(BaseParser):
    defaults = BaseParser.defaults.copy_and_update(
        regex=r'',
    )

    def __init__(self, *args, **kwargs):
        super(Regex, self).__init__(*args, **kwargs)
        self.re = re.compile(self.config.regex)

    def __str__(self):
        return 'REGEX-EXTRACTOR'

    def on_recv(self, data):
        match = self.re.search(data.message)
        if match:
            data.update_meta(match.groupdict())
        return self.output.send(data)


from __future__ import absolute_import, unicode_literals

from .base import BaseProcessor


class Stripper(BaseProcessor):

    def __str__(self):
        return 'STRIPPER'

    def on_recv(self, data):
        data.update_message(data.message.strip())
        return self.output.send(data)


# Default config.
# options/properties must be valid python identifiers, e.g. "default_pipe"
# names can be any string, e.g. "file->webui" (excluding ' ', '@')
from __future__ import absolute_import, unicode_literals


config = {
    # Sources format
    # YAML
    # sources:
    #   - (path | search pattern)
    #   # or
    #   # `handler`, `watcher`, `meta` are optional
    #   - (path | search pattern):
    #     handler: handler-name # default pipes.to-web
    #     watcher: watcher-name # default pollers.file-watcher
    #     meta: a-dictionary-containing-any-meta-info # e.g. {tags: [tag1, tag2]}
    #   # or
    #   - (path | search pattern): handler-name
    #   # or
    #   - (path | search pattern): {handler: pipes.to-web}
    #   # or
    #   - (path | search pattern): {watcher: poller.custom-file-poller}
    #   # or
    #   - (path | search pattern): {meta: {tags: [log]}}

    'sources': {
        '/var/log/*.log': {'handler': 'pipes.to-web'},
        '/var/log/*/*.log': {'handler': 'pipes.to-web'},
        '/var/log/syslog': 'pipes.to-web',
    },

    # Pipes format
    # YAML
    # pipes:
    #   pipe-name:
    #     - [namespace,...] pipe-object  # processors.stripper
    #     - ...                          # ns processors.stripper
    #     - ...config                    # ns1,ns2 processors.stripper

    'pipes': {
        'default': 'logdog.roles.pipes.Pipe',

        'to-web': [
            'watch processors.stripper',
            'watch connectors.zmq-tunnel@sender',
            'view connectors.zmq-tunnel@receiver',
            'view viewers.webui',
        ],

        'experiment-x001': {
            'cls': 'logdog.roles.pipes.Pipe',
            '*': [
                'watch processors.stripper',
                {'forwarders.broadcast': [
                    'watch viewers.console',
                    {'pipes.default': [
                        {'watch connectors.zmq-tunnel@sender': {'connect': 'tcp://localhost:7789'}},
                        {'view connectors.zmq-tunnel@receiver': {'bind': 'tcp://*:7789'}},
                        {'view forwarders.round-robin': [
                            'view viewers.webui',
                            'view viewers.null',
                        ]}
                    ]},
                ]},
            ],
        }
    },

    'options': {
        'sources': {
            'default_handler': 'pipes.to-web',
            'default_watcher': 'pollers.file-watcher',
            'index_file': '~/.config/logdog/sources-index.idx'
        }
    },

    # pollers
    'pollers': {
        'file-watcher': {
            'cls': 'logdog.roles.pollers.FileWatcher',
            'namespaces': ['watch'],
        },
    },

    # collectors
    'collectors': {},

    # processors / parsers / formatters
    'processors': {  # transform message
        'stripper': {'cls': 'logdog.roles.processors.Stripper'},
    },
    'parsers': {  # extract meta
        'regex': {'cls': 'logdog.roles.parsers.Regex', 'regex': r''},
    },
    'formatters': {  # add to meta formatted representation of message
        'formatter': {'cls': 'logdog.roles.formatters.Formatter'},
    },

    # forwarders
    'forwarders': {
        'broadcast': 'logdog.roles.forwarders.Broadcast',
        'round-robin': 'logdog.roles.forwarders.RoundRobin',
    },

    # connectors
    'connectors': {
        'zmq-tunnel': {
            'cls': 'logdog.roles.connectors.ZMQTunnel',
            '@sender': {'socket': 'PUSH', 'connect': ['tcp://localhost:45457']},
            '@receiver': {'socket': 'PULL', 'bind': ['tcp://*:45457']},
        }
    },

    # viewers
    'viewers': {
        'default': 'logdog.roles.viewers.Null',

        'webui': {
            'cls': 'logdog.roles.viewers.WebUI',
            'port': 8888,
            'debug': False,
        },
        'console': {
            'cls': 'logdog.roles.viewers.Console',
            'redirect_to': 'stdout',
        },
        'null': 'logdog.roles.viewers.Null',
    },

    # utils
    'utils': {
        'policies': {
            'growing-sleep': {
                'cls': 'logdog.core.policies.DefaultSleepPolicy'
            },
            'mostly-greedy': {
                'cls': 'logdog.core.policies.GreedyPolicy'
            }
        }
    }
}



from __future__ import absolute_import, unicode_literals

from tornado import gen

from .base import BaseForwarder


class RoundRobin(BaseForwarder):

    def __init__(self, *args, **kwargs):
        super(RoundRobin, self).__init__(*args, **kwargs)
        self._curr_target = 0

    def __str__(self):
        return 'ROUND-ROBIN:{}'.format(self.input)

    @gen.coroutine
    def _input_forwarder(self, data):
        if self.items:
            ret = self.items[self._curr_target].send(data)
            self._curr_target = (self._curr_target + 1) % len(self.items)
            if gen.is_future(ret):
                yield gen.maybe_future(ret)


from __future__ import absolute_import, unicode_literals

import logging
from tornado import gen

from logdog.core.base_role import BaseRole


logger = logging.getLogger(__name__)


class BaseForwarder(BaseRole):

    def __init__(self, *args, **kwargs):
        super(BaseForwarder, self).__init__(*args, **kwargs)
        self.items = [self.construct_subrole(name, conf) for name, conf in self.items]
        self.link_methods()
        for i in self.items:
            i.link_methods()

    def __str__(self):
        return 'FORWARDER:{}'.format(self._oid)

    @gen.coroutine
    def start(self):
        if not self.started:
            yield [i.start() for i in self.items]
            yield super(BaseForwarder, self).start()

    @gen.coroutine
    def stop(self):
        if self.started:
            yield [i.stop() for i in self.items]
            yield super(BaseForwarder, self).stop()




from __future__ import absolute_import, unicode_literals

from collections import deque
import logging
from tornado import gen

from logdog.core.base_role import BaseRole


logger = logging.getLogger(__name__)


class BaseCollector(BaseRole):
    defaults = BaseRole.defaults.inherit(
        buffer_size=1024,
    )

    def __init__(self, *args, **kwargs):
        super(BaseCollector, self).__init__(*args, **kwargs)
        self._buffer = deque(maxlen=self.config.buffer_size)
        self._wait_for_writable_buf = None

    def _check_buffer_renew_write_policy(self):
        write_available = len(self._buffer) <= self.config.buffer_size / 2
        if write_available and self._wait_for_writable_buf and not self._wait_for_writable_buf.done():
            self._wait_for_writable_buf.set_result(None)

    def _check_buffer_write_policy(self):
        return len(self._buffer) < self.config.buffer_size

    def _wait_for_buffer_writability(self):
        self._wait_for_writable_buf = gen.Future()
        return self._wait_for_writable_buf

    @gen.coroutine
    def on_recv(self, data):
        if not self._check_buffer_write_policy():
            logger.info('[%s] Buffer is full. Waiting until buffer is available...', self)
            yield self._wait_for_buffer_writability()
        self._buffer.append(data)

    def pop(self):
        try:
            data = self._buffer.popleft()
        except IndexError:
            return

        self._check_buffer_renew_write_policy()
        return data




from __future__ import absolute_import, unicode_literals

import logging

from logdog.core.base_role import BaseRole


logger = logging.getLogger(__name__)


class BaseViewer(BaseRole):

    def on_recv(self, data):
        logger.debug(data)


from logdog.core.utils.six import *


__all__ = ['import_object', 'text_type']


# tornado import_object
from tornado.util import import_object as _import_object
if PY2:
    import_object = lambda name: _import_object(str(name))
if PY3:
    import_object = _import_object



from __future__ import absolute_import, unicode_literals

import copy
import logging
import os

from tornado.util import ObjectDict
from yaml import load
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

from logdog.default_config import config as logdog_default_config
from logdog.compat import text_type, import_object


logger = logging.getLogger(__name__)


class ConfigError(Exception):
    pass

_unset = object()


def handle_as_list(conf):
    if isinstance(conf, dict):
        return conf.items()
    if isinstance(conf, (list, tuple)):
        l = []
        for item in conf:
            if isinstance(item, basestring):
                l.append((item, None))
            elif isinstance(item, dict) and len(item) == 1:
                l.append(item.items()[0])
            elif isinstance(item, (list, tuple)) and item:
                l.append((item[0], item[1:] if len(item) > 2 else item[1]))
            else:
                l.append((item, None))
        return l
    return [(conf, None)]


def is_importable(key):
    return key in ('cls',)


class Config(ObjectDict):
    namespace_delimiter = ' '
    namespace_default = '*'
    subconfig_namespace_delimiter = '@'

    _classes_cache = {}
    _conf_path_cache = {}

    def copy(self):
        return copy.copy(self)

    def deepcopy(self):
        return copy.deepcopy(self)

    def copy_and_update(self, *args, **config):
        conf = self.deepcopy()
        for a in args:
            if isinstance(a, dict):
                conf.update(a)
        conf.update(config)
        return conf

    __call__ = copy_and_update

    def walk_conf(self, path, default=_unset):
        _cache = self._conf_path_cache.get(path)
        if _cache is not None:
            return copy.deepcopy(_cache)

        target = self
        for part in path.split('.'):
            try:
                target = target[part]
            except (KeyError, IndexError):
                if default is not _unset:
                    return default
                raise

        self._conf_path_cache[path] = target
        return copy.deepcopy(target)

    def find_conf(self, name, fallback=None):
        name = name.strip().lstrip(self.namespace_delimiter).rstrip(self.subconfig_namespace_delimiter)

        namespace = self.namespace_default
        if self.namespace_delimiter in name:
            namespace, tmp, name = name.partition(self.namespace_delimiter)

        extra_conf = None
        if self.subconfig_namespace_delimiter in name:
            name, tmp, extra_conf = name.rpartition(self.subconfig_namespace_delimiter)
            extra_conf_relative = '{}.{}{}'.format(name, self.subconfig_namespace_delimiter, extra_conf)
            extra_conf = self.walk_conf(extra_conf_relative, default=None)

        try:
            target = self.walk_conf(name)
        except KeyError:
            if fallback:
                return self.find_conf(name=fallback)
            else:
                raise

        if isinstance(target, (list, tuple)):
            target = handle_as_list(target)

        elif extra_conf and isinstance(target, dict) and isinstance(extra_conf, dict):
            target.update(extra_conf)

        return target, name, namespace.split(',')

    def find_class(self, name, fallback=None):
        initial_name = name
        _cache = self._classes_cache.get(initial_name)
        if _cache is not None:
            return copy.deepcopy(_cache)

        conf, name, namespaces = self.find_conf(name, fallback=None)
        if isinstance(conf, basestring):
            conf = {'cls': conf}
        elif isinstance(conf, (list, tuple)):
            conf = {'*': conf}

        try:
            cls = conf.pop('cls')
        except KeyError:
            if name != 'default':
                fallback = fallback if fallback else ''.join(name.rpartition('.')[:-1] + ('default',))
                cls, f_name, f_conf, f_ns = self.find_class(name=fallback)
                f_conf.update(conf)
                conf = f_conf
                namespaces = list(set(namespaces).union(f_ns))
            else:
                raise

        # sanitize conf
        for k in conf.keys():
            if k.startswith(self.subconfig_namespace_delimiter):
                conf.pop(k)

        if isinstance(cls, basestring):
            try:
                cls = import_object(cls)
            except ImportError:
                logger.warning('Could not import "%s".', cls)

        if 'namespaces' in conf:
            if tuple(namespaces) == (self.namespace_default,):
                namespaces = tuple(conf['namespaces'])
            else:
                namespaces = tuple(set(namespaces).union(conf['namespaces']))

        self._classes_cache[initial_name] = cls, name, conf, tuple(namespaces)
        return cls, name, copy.deepcopy(conf), tuple(namespaces)

    def find_and_construct_class(self, name, fallback=None, args=(), kwargs=None):
        found_cls, name, defaults, ns = self.find_class(name=name, fallback=fallback)

        if kwargs:
            defaults.update(kwargs)

        kw = defaults
        kw.update(kw.pop('**', {}))
        kw['namespaces'] = set(kw.get('namespaces', ())).union(ns)
        kw['config_name'] = name

        args = kw.pop('*', args)

        if hasattr(found_cls, 'factory'):
            found_cls = found_cls.factory

        args = handle_as_list(args)

        if callable(found_cls):
            return found_cls(*args, **kw)
        raise ConfigError('{} is not a class. Cannot be imported or called.'.format(found_cls))


class ConfigLoader(object):

    def __init__(self, path):
        self._path = path

    def _load_user_config(self):
        if not os.path.exists(self._path):
            raise ConfigError('[{}] Config file does not exist.'.format(self._path))
        with open(self._path, 'r') as f:
            return load(f, Loader=Loader)

    def _merge_configs(self, config, updater):
        if isinstance(config, dict):
            if isinstance(updater, dict):
                missing_keys = set(updater).difference(config)
                existing_keys = set(updater).intersection(config)

                cfg = config.copy()
                for k in missing_keys:
                    cfg[k] = updater[k]

                for k in existing_keys:
                    cfg[k] = self._merge_configs(cfg[k], updater[k])

                return cfg

        return updater

    def load_config(self, default_only=False):
        user_config = self._load_user_config() if not default_only else Config()
        default_config = copy.deepcopy(logdog_default_config)

        if not isinstance(user_config, dict):
            raise ConfigError('[{}] Invalid config file: must have \"key: value\" format.'.format(self._path))

        def walk(cfg, key=None):
            if isinstance(cfg, dict):
                return Config((k, walk(v, key=k)) for k, v in cfg.items())
            if isinstance(cfg, (tuple, list)):
                return map(walk, cfg)
            if isinstance(cfg, (str, text_type)):
                ret = cfg.format(default=default_config)
                if is_importable(key):
                    try:
                        ret = import_object(ret)
                    except ImportError:
                        logger.debug('[%s] Faced non-importable path: \"%s: %s\".', self._path, key, ret)
                return ret
            return cfg

        default_config = walk(default_config)
        user_config = walk(user_config)

        conf = self._merge_configs(default_config, user_config)
        conf.sources = handle_as_list(conf.sources)

        return conf


from __future__ import absolute_import, unicode_literals

from tornado import gen

from .base import BaseForwarder


class Broadcast(BaseForwarder):
    def __str__(self):
        return 'BROADCAST:{}'.format(self.input)

    @gen.coroutine
    def _input_forwarder(self, data):
        return [gen.maybe_future(i.send(data)) for i in self.items]


#!/usr/bin/env python
import os
from setuptools import find_packages, setup

from logdog import (
    __version__ as version,
    __author__ as author,
    __email__ as author_email)


requires = [
    r.strip() for r in
    open(os.path.join(os.getcwd(), 'requirements.txt')).readlines()
]

setup(
    name='logdog',
    version=version,
    description='logs watching + webui',
    author=author,
    author_email=author_email,
    url='https://github.com/miphreal/python-logdog/',
    packages=find_packages(exclude=['tests']),
    scripts=['scripts/logdog'],
    install_requires=requires,
    package_data={'logdog': ['roles/viewers/webui/assets/html/index.html',
                             'roles/viewers/webui/static/img/icons/navigation/svg/production/ic_menu_18px.svg',
                             'roles/viewers/webui/static/css/app.css',
                             'roles/viewers/webui/static/js/app.js']}
)


from __future__ import absolute_import, unicode_literals

from .base import BaseFormatter


class Formatter(BaseFormatter):
    defaults = BaseFormatter.defaults.copy_and_update(
        format='{msg.source!s}: {msg.message!s}',
        format_meta_name=None,  # None means 'replace msg.message'
    )

    def __init__(self, *args, **kwargs):
        super(Formatter, self).__init__(*args, **kwargs)
        self.format = self.config.format
        self.format_meta_name = self.config.format_meta_name

    def __str__(self):
        return 'FORMATTER'

    def on_recv(self, data):
        msg = self.format.format(msg=data)
        if self.format_meta_name is not None:
            data.update_meta({self.format_meta_name: msg})
        else:
            data.update_message(msg)
        return self.output.send(data)